﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Drawing;

namespace KV6ToPNG
{
    public class PointCloud
    {
        protected int MaxSize { get { return Math.Max(XSize, Math.Max(YSize, ZSize)); } }
        public int XSize { get; set; }
        public int YSize { get; set; }
        public int ZSize { get; set; }

        public Point[, ,] Data { get; set; }

        #region Load KV6
        private class voxsub
        {
            public int r, g, b, dummy, height, visibilityFlag, normalIndex;
        };
        public static PointCloud LoadFormKV6(String Filename)
        {
            PointCloud PC = new PointCloud();

            BinaryReader br = new BinaryReader(File.Open(Filename, FileMode.Open));
            int intro = br.ReadInt32();
            if (intro != 0x6c78764b)
            {
                br.Close();
                return null;
            }

            PC.XSize = br.ReadInt32();
            PC.YSize = br.ReadInt32();
            PC.ZSize = br.ReadInt32();
            PC.Data = new Point[PC.XSize, PC.YSize, PC.ZSize];

            //pivot point (not important)
            br.ReadInt32();
            br.ReadInt32();
            br.ReadInt32();

            int surfaceCount = br.ReadInt32();
            

            voxsub[] voxes = new voxsub[surfaceCount];

            for (int itr = 0; itr < surfaceCount; itr++)
            {
                voxes[itr] = new voxsub();
                voxes[itr].r = br.ReadByte();
                voxes[itr].g = br.ReadByte();
                voxes[itr].b = br.ReadByte();

                voxes[itr].dummy = br.ReadByte();
                if (voxes[itr].dummy != 128)
                {
                    br.Close();
                    return null;
                }

                voxes[itr].height = br.ReadByte();
                voxes[itr].height += br.ReadByte() * 256;//shift
                voxes[itr].visibilityFlag = br.ReadByte();
                voxes[itr].normalIndex = br.ReadByte();
            }


            for (int xp = 0; xp < PC.XSize; xp++)
            {
                br.ReadInt32();//theoreticaly not imporant
            }

            //Number of surface voxels present in column (x,y)
            int voxCount;
            int voxID = 0;
            for (int xp = 0; xp < PC.XSize; xp++)
            {
                for (int yp = 0; yp < PC.YSize; yp++)
                {
                    voxCount = br.ReadInt16();
                    for (int itr = 0; itr < voxCount; itr++)
                    {
                        Point pnt = new Point(voxes[voxID].r, voxes[voxID].g, voxes[voxID].b, voxes[voxID].visibilityFlag, xp, yp, voxes[voxID].height);
                        pnt.VoxNormal = voxes[voxID].normalIndex;
                        PC.WriteVoxelAndAir(xp, yp, voxes[voxID].height, pnt);
                        voxID++;
                    }
                }
            }

            br.Close();

            return PC;
        }

        private void WriteVoxelAndAir(int x, int y, int z, Point point)
        {
            Data[x, y, z] = point;
            //SetData(x, y, z, point);

            if (x > 0 && (point.SurfaceFlag & 0x1) != 0)
                Data[x - 1, y, z] = new Point();
            if (x < XSize - 1 && (point.SurfaceFlag & 0x2) != 0)
                Data[x + 1, y, z] = new Point();

            if (y > 0 && (point.SurfaceFlag & 0x4) != 0)
                Data[x, y - 1, z] = new Point();
            if (y != YSize - 1 && (point.SurfaceFlag & 0x8) != 0)
                Data[x, y + 1, z] = new Point();

            if (z > 0 && (point.SurfaceFlag & 0x10) != 0)
                Data[x, y, z - 1] = new Point();
            if (z != ZSize - 1 && (point.SurfaceFlag & 0x20) != 0)
                Data[x, y, z + 1] = new Point();
/*            if (x > 0 && (point.SurfaceFlag & 0x1) != 0)
                SetData(x - 1, y, z,new Point());
            if (x < XSize - 1 && (point.SurfaceFlag & 0x2) != 0)
                SetData(x + 1, y, z, new Point());

            if (y > 0 && (point.SurfaceFlag & 0x4) != 0)
                SetData(x, y - 1, z, new Point());
            if (y != YSize - 1 && (point.SurfaceFlag & 0x8) != 0)
                SetData(x, y + 1, z, new Point());

            if (z > 0 && (point.SurfaceFlag & 0x10) != 0)
                SetData(x, y, z - 1, new Point());
            if (z != ZSize - 1 && (point.SurfaceFlag & 0x20) != 0)
                SetData(x, y, z + 1, new Point());*/
        }

        void SetData(int x, int y, int z, Point point)
        {
            if (Data[x, y, z] != null && !Data[x, y, z].IsAir)
                return;
            Data[x, y, z] = point;
        }

        #endregion Load KV6

        #region Save KV6
        protected Point[] GetSurfaces()
        {
            List<Point> PointList = new List<Point>();
            for (int ix = 0; ix < XSize; ix++)
                for (int iy = 0; iy < YSize; iy++)
                    for (int iz = 0; iz < ZSize; iz++)
                    {
                        if (Data[ix, iy, iz] != null && Data[ix, iy, iz].IsEdge)
                            PointList.Add(Data[ix, iy, iz]);
                    }
            return PointList.ToArray();
        }

        public void SaveToKV6(String Filename)
        {
            BinaryWriter bw = new BinaryWriter(File.Open(Filename, FileMode.OpenOrCreate));

            bw.Write((UInt32)0x6c78764b);

            bw.Write((Int32)XSize);
            bw.Write((Int32)YSize);
            bw.Write((Int32)ZSize);

            //pivot point
            bw.Write((Int32)1099956224);
            bw.Write((Int32)1102577664);
            bw.Write((Int32)1108869120);

            Point[] SurfacePointList = GetSurfaces();

            bw.Write((Int32)SurfacePointList.Length);

            for (int itr = 0; itr < SurfacePointList.Length; itr++)
            {
                int RGBA = SurfacePointList[itr].RGBA;
                int r = RGBA & 0x000000ff, g = (RGBA & 0x0000ff00) >> 8, b = (RGBA & 0x00ff0000) >> 16;
                bw.Write((byte)r);
                bw.Write((byte)g);
                bw.Write((byte)b);

                bw.Write((byte)128);

                int zPos = SurfacePointList[itr].z;
                bw.Write((byte)(zPos & 0x00ff));
                bw.Write((byte)((zPos & 0xff00)>> 8));

                bw.Write((byte)(SurfacePointList[itr].SurfaceFlag & 0x3f));
                bw.Write((byte)SurfacePointList[itr].VoxNormal);//fixme normal
            }

            int surfaceCount = 0;
            for (int xp = 0; xp < XSize; xp++)
            {
                for (int itr = 0; itr < SurfacePointList.Length; itr++)
                {
                    if (SurfacePointList[itr].x == xp)
                        surfaceCount++;
                }
                bw.Write((Int32)surfaceCount);
                surfaceCount = 0;
            }

            surfaceCount = 0;
            for (int xp = 0; xp < XSize; xp++)
                for (int yp = 0; yp < YSize; yp++)
                {
                    for (int itr = 0; itr < SurfacePointList.Length; itr++)
                    {
                        if (SurfacePointList[itr].x == xp && SurfacePointList[itr].y == yp)
                            surfaceCount++;
                    }
                    bw.Write((Int16)surfaceCount);
                    surfaceCount = 0;
                }

            bw.Close();
        }
        #endregion Save KV6

        #region Fill uninitialized point
        private Point FindNearest(int initX, int initY, int initZ, int Offset)
        {
            int step = -1;
            int x, y, z;
            int Largest = MaxSize * 2;
            for (int itr = 0; itr < MaxSize * 2; itr++)
            {
                if ((Offset & 0x8) != 0)
                {
                    x = initX + (step);
                    y = initY + (step);
                    z = initZ + (step);
                }
                else
                {
                    x = initX - (step);
                    y = initY - (step);
                    z = initZ - (step);
                }

                int of2 = Offset % 6;
                if (of2 == 0)
                {
                    if (x < 0)
                        return new Point();
                    else if (x > (XSize-1))
                        return new Point();
                    else if (Data[x, initY, initZ] != null)
                        return new Point(Data[x, initY, initZ].RGBA, 0x80, x, initY, initZ);//solid point

                    if (y < 0)
                        return new Point();
                    else if (y > (YSize - 1))
                        return new Point();
                    else if (Data[initX, y, initZ] != null)
                        return new Point(Data[initX, y, initZ].RGBA, 0x80, initX, y, initZ);

                    if (z < 0)
                        return new Point();
                    else if (z > (ZSize - 1))
                        return new Point();
                    else if (Data[initX, initY, z] != null)
                        return new Point(Data[initX, initY, z].RGBA, 0x80, initX, initY, z);
                }
                if (of2 == 1)
                {
                    if (x < 0)
                        return new Point();
                    else if (x > (XSize - 1))
                        return new Point();
                    else if (Data[x, initY, initZ] != null)
                        return new Point(Data[x, initY, initZ].RGBA, 0x80, x, initY, initZ);

                    if (z < 0)
                        return new Point();
                    else if (z > (ZSize - 1))
                        return new Point();
                    else if (Data[initX, initY, z] != null)
                        return new Point(Data[initX, initY, z].RGBA, 0x80, initX, initY, z);

                    if (y < 0)
                        return new Point();
                    else if (y > (YSize - 1))
                        return new Point();
                    else if (Data[initX, y, initZ] != null)
                        return new Point(Data[initX, y, initZ].RGBA, 0x80, initX, y, initZ);

                }
                if (of2 == 2)
                {
                    if (y < 0)
                        return new Point();
                    else if (y > (YSize - 1))
                        return new Point();
                    else if (Data[initX, y, initZ] != null)
                        return new Point(Data[initX, y, initZ].RGBA, 0x80, initX, y, initZ);

                    if (x < 0)
                        return new Point();
                    else if (x > (XSize - 1))
                        return new Point();
                    else if (Data[x, initY, initZ] != null)
                        return new Point(Data[x, initY, initZ].RGBA, 0x80, x, initY, initZ);

                    if (z < 0)
                        return new Point();
                    else if (z > (ZSize - 1))
                        return new Point();
                    else if (Data[initX, initY, z] != null)
                        return new Point(Data[initX, initY, z].RGBA, 0x80, initX, initY, z);
                }
                if (of2 == 3)
                {
                    if (y < 0)
                        return new Point();
                    else if (y > (YSize - 1))
                        return new Point();
                    else if (Data[initX, y, initZ] != null)
                        return new Point(Data[initX, y, initZ].RGBA, 0x80, initX, y, initZ);

                    if (z < 0)
                        return new Point();
                    else if (z > (ZSize - 1))
                        return new Point();
                    else if (Data[initX, initY, z] != null)
                        return new Point(Data[initX, initY, z].RGBA, 0x80, initX, initY, z);
                    if (x < 0)
                        return new Point();
                    else if (x > (XSize - 1))
                        return new Point();
                    else if (Data[x, initY, initZ] != null)
                        return new Point(Data[x, initY, initZ].RGBA, 0x80, x, initY, initZ);

                }
                if (of2 == 4)
                {
                    if (z < 0)
                        return new Point();
                    else if (z > (ZSize - 1))
                        return new Point();
                    else if (Data[initX, initY, z] != null)
                        return new Point(Data[initX, initY, z].RGBA, 0x80, initX, initY, z);

                    if (x < 0)
                        return new Point();
                    else if (x > (XSize - 1))
                        return new Point();
                    else if (Data[x, initY, initZ] != null)
                        return new Point(Data[x, initY, initZ].RGBA, 0x80, x, initY, initZ);

                    if (y < 0)
                        return new Point();
                    else if (y > (YSize - 1))
                        return new Point();
                    else if (Data[initX, y, initZ] != null)
                        return new Point(Data[initX, y, initZ].RGBA, 0x80, initX, y, initZ);
                }
                if (of2 == 5)
                {
                    if (z < 0)
                        return new Point();
                    else if (z > (ZSize - 1))
                        return new Point();
                    else if (Data[initX, initY, z] != null)
                        return new Point(Data[initX, initY, z].RGBA, 0x80, initX, initY, z);

                    if (y < 0)
                        return new Point();
                    else if (y > (YSize - 1))
                        return new Point();
                    else if (Data[initX, y, initZ] != null)
                        return new Point(Data[initX, y, initZ].RGBA, 0x80, initX, y, initZ);

                    if (x < 0)
                        return new Point();
                    else if (x > (XSize - 1))
                        return new Point();
                    else if (Data[x, initY, initZ] != null)
                        return new Point(Data[x, initY, initZ].RGBA, 0x80, x, initY, initZ);

                }

                step = (~step) + (step > 0 ? 0 : 1);
            }

            return null;
        }

        public void FillPoint()
        {
            int ix, iy, iz;
            UInt64 step;
            uint itr = 0;
            uint XSize = (uint)this.XSize;
            uint YSize = (uint)this.YSize;
            uint ZSize = (uint)this.ZSize;
            uint MaxSize = (uint)XSize * (uint)YSize * (uint)ZSize;

            for (; itr < XSize * YSize * ZSize; itr++)
            {//262144 0x3ffff
                step = (UInt64)95731 + itr * (UInt64)104729 * 104723;// (UInt64)1148487369611039;
                step = step % MaxSize;
                ix = (int)(step % XSize);
                iy = (int)((step / XSize) % YSize);
                iz = (int)((step / (XSize * YSize)) % ZSize);

                if (Data[ix, iy, iz] == null)
                    Data[ix, iy, iz] = FindNearest(ix, iy, iz, (int)itr);
            }
            itr = 0;
        }
        #endregion Fill uninitialized point

        #region GetSlice
        public Slice GetSlice(SliceAxis Axis, int Layer)
        {
            Slice slice = new Slice();
            slice.Axis = Axis;
            slice.Layer = Layer;
            slice.Rings = new List<Ring>();

            if (Axis == SliceAxis.X_Slice)
            {
                slice.Data = new Point[YSize, ZSize];

                for (int yi = 0; yi < YSize; yi++)
                    for (int zi = 0; zi < ZSize; zi++)
                    {
                        slice.Data[yi, zi] = new Point(Data[Layer, yi, zi]);
                    }
            }
            else if (Axis == SliceAxis.Y_Slice)
            {
                slice.Data = new Point[XSize, ZSize];

                for (int xi = 0; xi < XSize; xi++)
                    for (int zi = 0; zi < ZSize; zi++)
                    {
                        slice.Data[xi, zi] = new Point(Data[xi, Layer, zi]);
                    }
            }
            else if (Axis == SliceAxis.Z_Slice)
            {
                slice.Data = new Point[XSize, YSize];

                for (int xi = 0; xi < XSize; xi++)
                    for (int yi = 0; yi < YSize; yi++)
                    {
                        slice.Data[xi, yi] = new Point(Data[xi, yi, Layer]);
                    }
            }

            return slice;
        }
        #endregion GetSlice

        #region Export BMP
        public Bitmap ExportToBitmap()
        {
            int Scale = (int)Math.Ceiling(Math.Sqrt(ZSize));
            Bitmap bmp = new Bitmap(XSize * Scale, YSize * Scale);
            for (int ix = 0; ix < XSize; ix++)
                for (int iy = 0; iy < YSize; iy++)
                    for (int iz = 0; iz < ZSize; iz++)
                    {
                        if (Data[ix, iy, iz] != null)
                        {
                            int bx = ix + (iz % Scale) * XSize;
                            int by = iy + (iz / Scale) * YSize;

                            if (Data[ix, iy, iz] != null)
                                bmp.SetPixel(bx, by, Color.FromArgb(Data[ix, iy, iz].RGBA));
                        }
                    }

            return bmp;
        }
        #endregion Export BMP

        #region Import BMP as PointCloud
        public static PointCloud ImportBitmap(Bitmap bmp, int XSize, int YSize, int ZSize)
        {
            PointCloud PC = new PointCloud();
            PC.XSize = XSize;
            PC.YSize = YSize;
            PC.ZSize = ZSize;
            PC.Data = new Point[XSize, YSize, ZSize];
            int Scale = (int)Math.Ceiling(Math.Sqrt(ZSize));
            for (int ix = 0; ix < XSize; ix++)
                for (int iy = 0; iy < YSize; iy++)
                    for (int iz = 0; iz < ZSize; iz++)
                    {
                        int bx = ix + (iz % Scale) * XSize;
                        int by = iy + (iz / Scale) * YSize;
                        Color col = bmp.GetPixel(bx, by);

                        Point pnt = new Point(col.ToArgb(), 0x80, ix, iy, iz);
                        if ((pnt.RGBA & 0xff000000) == 0x00000000)//air point
                            pnt.SurfaceFlag = 0x40;
                        PC.Data[ix, iy, iz] = pnt;
                    }

            for (int ix = 0; ix < XSize; ix++)
                for (int iy = 0; iy < YSize; iy++)
                    for (int iz = 0; iz < ZSize; iz++)
                        if (PC.Data[ix, iy, iz].IsSolid)
                        {
                            int Surface = 0x0;

                            if (ix == 0 || PC.Data[ix - 1, iy, iz].IsAir)
                                Surface = Surface | 0x01;
                            if (ix == XSize - 1 || PC.Data[ix + 1, iy, iz].IsAir)
                                Surface = Surface | 0x02;

                            if (iy == 0 || PC.Data[ix, iy - 1, iz].IsAir)
                                Surface = Surface | 0x04;
                            if (iy == YSize - 1 || PC.Data[ix, iy + 1, iz].IsAir)
                                Surface = Surface | 0x08;

                            if (iz == 0 || PC.Data[ix, iy, iz - 1].IsAir)
                                Surface = Surface | 0x10;
                            if (iz == ZSize - 1 || PC.Data[ix, iy, iz + 1].IsAir)
                                Surface = Surface | 0x20;

                            if (Surface != 0)
                                PC.Data[ix, iy, iz].SurfaceFlag = Surface;
                        }

            return PC;
        }
        #endregion Import BMP as PointCloud
    }
}
